/* 
 * File:   CHLAAlignment.cpp
 * Author: mwittig
 * 
 * Created on January 24, 2013, 4:58 PM
 */
#include <string>
#include <iostream>
#include <stdio.h>
#include <stdlib.h>
#include <sstream>
#include <map>
#include <fstream>
#include <vector>
#include <set>
#include <iomanip>
#include <algorithm>
#include <utility>
#include <numeric>
#include <math.h>
#include <assert.h>

#include "CBlat.h"
#include "CHLAAlignment.h"

#define BOOST_ALL_NO_LIB

CHLAAlignment::CHLAAlignment()
{
    m_read_length = 100;
}

CHLAAlignment::CHLAAlignment(const string& seq)
{
    assert("Achtung verwendung dieses Konstruktors nur bei Arbeiten ohne reads!");
    m_read_length = 100;
    getExonBlocks(seq);
}

CHLAAlignment::CHLAAlignment(const string& seq, const unsigned int& read_length)
{
    m_read_length = read_length;
    getExonBlocks(seq);
}

CHLAAlignment::CHLAAlignment(const CHLAAlignment& orig)
{
    m_exon_coordinates = orig.m_exon_coordinates;
    m_exon_sequences   = orig.m_exon_sequences;
    m_aligned_reads    = orig.m_aligned_reads;
    m_coverage_metrics = orig.m_coverage_metrics;
    m_read_length      = orig.m_read_length;
}

CHLAAlignment::~CHLAAlignment()
{

}

/*
 * concateniert alle exone, die die Mindestlänge besitzen
 */
string CHLAAlignment::getAlignableSequence()
{
    string strRet("");

    for(vector<string>::iterator iter =  m_exon_sequences.begin();iter !=  m_exon_sequences.end();iter++)
        if(iter->size() >= _CHLAALIGNMENT_MIN_EXON_LENGTH)
            strRet+=*iter;

    // cout << strRet << endl;
    return strRet;
}

long   CHLAAlignment::getAlignableExonCount()
{
    long lngRet = 0;

    for(vector<string>::iterator iter =  m_exon_sequences.begin();iter !=  m_exon_sequences.end();iter++)
        if(iter->size() >= _CHLAALIGNMENT_MIN_EXON_LENGTH)
            lngRet++;

    return lngRet;
}

/*
 *  Liefert die Start und Endpunkte der einzelnen Exons
 *  Exons sind durch | getrennt. Die Funktion sagt aber alles,
 *  was nicht ACGTacgt ist, ist ein Trenner
 *
 *  Rückgabe ist ein pair mit Start-Idx und End-Idx der Exonsequenz
 *  Start Idx ist 0-basiert, end-idx ist 1-basiert
 */
void CHLAAlignment::getExonBlocks(const string& seq)
{
    m_exon_sequences = Tokenize(seq,"|");
    int j=0;
    for(int i = 0; i < m_exon_sequences.size(); i++)
    {
        pair<unsigned int, unsigned int> act_coord(j,j+m_exon_sequences[i].size());
        m_exon_coordinates.push_back(act_coord);
        j+=m_exon_sequences[i].size();
    }
}

vector<string> CHLAAlignment::Tokenize(const string& str, const string& delimiters)
{
    vector<string> tokens;
    // Skip delimiters at beginning.
    string::size_type lastPos = str.find_first_not_of(delimiters, 0);
    // Find first "non-delimiter".
    string::size_type pos     = str.find_first_of(delimiters, lastPos);

    while (string::npos != pos || string::npos != lastPos)
    {
        // Found a token, add it to the vector.
        tokens.push_back(str.substr(lastPos, pos - lastPos));
        // Skip delimiters.  Note the "not_of"
        if(str.length() > pos && delimiters.find(str[pos]) != string::npos )
            lastPos=pos+1;
        else
            lastPos = str.find_first_not_of(delimiters, pos);
        // Find next "non-delimiter"
        pos = str.find_first_of(delimiters, lastPos);
    }
    return tokens;
}

string CHLAAlignment::getExonReport()
{
    ostringstream strRet("");

    strRet << setw(3) << right << m_exon_sequences.size() << " exons" << endl
           << " following coordinates are 0-based for start and 1-based for end coordinates http://genome.ucsc.edu/FAQ/FAQtracks.html#tracks1" << endl;
    for(int i = 0; i < m_exon_coordinates.size(); i++)
        strRet << " exon " << setw(3) << right << i << " from " << setw(5) << right << m_exon_coordinates[i].first << " to " << setw(5) << right << m_exon_coordinates[i].second << endl;

    for(int i = 0; i < m_exon_sequences.size(); i++)
        strRet << " exon " << setw(3) << right << i << ": " <<  m_exon_sequences[i] << endl;

    return strRet.str();
}

string CHLAAlignment::getExonsAsFasta()
{
    ostringstream strRet("");

    for(int i = 0; i < m_exon_sequences.size(); i++)
        strRet << '>' << i << endl <<  m_exon_sequences[i] << endl;

    return strRet.str();
}

void CHLAAlignment::getExonsAsFasta(const string& filename)
{
    try
    {
        ofstream out(filename.c_str());
        out << getExonsAsFasta();
        out.close();
    }
    catch(...)
    {
        throw(string("Can't open ")+filename);
    }
}
void CHLAAlignment::transferReadsFromBlat(const multimap<string,CBlat::stcAlignment>&  aligned_reads)
{
    vector<unsigned int> alignment_offset(m_exon_sequences.size(),0);

    for(int i = 1; i < alignment_offset.size(); i++)
        alignment_offset[i] = alignment_offset[i-1] + m_exon_sequences[i-1].size();

    for(multimap<string,CBlat::stcAlignment>::const_iterator iter = aligned_reads.begin(); iter != aligned_reads.end(); iter++)
    {
        stcAlignment act_align;
        act_align.strand      = iter->second.strand;
        act_align.AlignLength = iter->second.AlignLength;
        act_align.ReadStart   = iter->second.ReadStart;
        act_align.RefStart    = alignment_offset[iter->second.RefId] + iter->second.RefStart;
        map<string,stcAlignment>::iterator iter_old = m_aligned_reads.find(iter->first);
        if(iter_old == m_aligned_reads.end())
            m_aligned_reads[iter->first]=act_align; // unbekannter read wird gespeichert
        else // gibt es den read schon, dann nimm den besseren oder einen zufälligen
        {
            if(iter_old->second.AlignLength == act_align.AlignLength )
            {
                if(0 == ( rand() % 2 ) )
                     m_aligned_reads[iter->first]=act_align;
            }

        }
    }
}

vector<pair<long,long> > CHLAAlignment::getSortedReads(bool zero_based_end)
{
    vector<pair<long,long> > vctReturn;
    for(map<string,stcAlignment>::const_iterator iter = m_aligned_reads.begin(); iter != m_aligned_reads.end(); iter++)
    {
        if(zero_based_end)
            vctReturn.push_back(pair<long,long>(iter->second.RefStart,iter->second.RefStart+iter->second.AlignLength-1));
        else
            vctReturn.push_back(pair<long,long>(iter->second.RefStart,iter->second.RefStart+iter->second.AlignLength));
    }
    sort(vctReturn.begin(),vctReturn.end(),sort_alignments);
    return vctReturn;
}

void CHLAAlignment::getSortedReads(const string& filename, bool zero_based_end)
{
    int k = 0;
    ofstream out(filename.c_str());
    if(!out)
        throw(string("Can't open ")+filename);

    vector<pair<long,long> > sorted_reads = getSortedReads(zero_based_end);
    for(vector<pair<long,long> >::iterator iter = sorted_reads.begin(); iter != sorted_reads.end(); iter++)
    {
        out << iter->first << '\t' << k << '\t' << iter->second << '\t' << k << endl;
        if(++k==101)
            k=0;
    }
    out.close();
}


bool CHLAAlignment::sort_alignments(const pair<long,long>& one, const pair<long,long>& two)
{
    if(one.first < two.first)
        return true;
    else if(one.first > two.first)
        return false;
    else if(one.second < two.second)
        return true;
    return false;
}

double CHLAAlignment::getAreaUnderCurveError()
{
    vector<long> ideal = getIdealCoverage();
    vector<long> real  = getCoverage();

    if(ideal.size() != real.size())
        throw("Something went wrong with the area under curve calculation");

    double ret_value = 0.0f;
    size_t not_coverable = 0;
    for(size_t i = 0; i < ideal.size(); i++)
    {
        if(ideal[i] == 0)
            not_coverable++;
        else
            ret_value+=ideal[i]-real[i];
    }

    return ret_value/(ideal.size()-not_coverable);

}

vector<double> CHLAAlignment::getAreaUnderCurveInterpolator()
{
    try
    {
        vector<long> ideal = getIdealCoverage();
        vector<long> real  = getCoverage();

        vector<double>  vctRet;

        if(ideal.size() != real.size())
            throw("Something went wrong with the area under curve calculation");
        int k = 0;
        for(int i = 0; i < m_exon_sequences.size(); i++)
        {
            double cum_ideal = 0.0f;
            double cum_real  = 0.0f;
            for(int j = 0; j < m_exon_sequences[i].size(); j++,k++)
            {
                cum_ideal+=ideal[k];
                cum_real+=real[k];
            }
            if(cum_ideal == 0.0f || cum_real == 0.0f)
                vctRet.push_back(1.0f);
            else
                vctRet.push_back(cum_ideal/cum_real);
        }
        return vctRet;
    }
    catch(...)
    {
        cerr << "unexpected error in vector<double> CHLAAlignment::getAreaUnderCurveInterpolator(). Very likely out of bounds error ..." << endl;
        exit (EXIT_FAILURE);
    }
}

vector<long> CHLAAlignment::getCoverage()
{
    unsigned int k = getcDNAsize();
    vector<long> vctReturn(k,0);

    for(map<string,stcAlignment>::const_iterator iter = m_aligned_reads.begin(); iter != m_aligned_reads.end(); iter++)
        for(int i = iter->second.RefStart; i < iter->second.RefStart + iter->second.AlignLength; i++)
            vctReturn[i]++;

    return vctReturn;
}

vector<long> CHLAAlignment::getIdealCoverage()
{
    unsigned int k = getcDNAsize();

    vector<long> vctReturn(k,0);

    k = 0;
    for(int i = 0; i < m_exon_sequences.size(); i++)
    {
        if(m_exon_sequences[i].size() >= 30)
        {
            int alignment_length_treshold=m_exon_sequences[i].size();
            if(m_exon_sequences[i].size() >= 70)
                alignment_length_treshold = 70;

            vector<long> vctTmp(m_exon_sequences[i].size()+m_read_length-alignment_length_treshold,0);
            for(size_t j = 0; j <= vctTmp.size()-alignment_length_treshold; j++)
                for(size_t l = 0; l < m_read_length && l+j < vctTmp.size(); l++)
                    vctTmp[l+j]++;
            vctTmp.erase(vctTmp.begin(),vctTmp.begin()+(m_read_length-alignment_length_treshold));
            // cout << vctTmp.size() << " - " << m_exon_sequences[i].size() << endl;
            for(size_t j = 0; j < vctTmp.size();j++)
                vctReturn[j+k]=vctTmp[j];
        }
        k+=m_exon_sequences[i].size();
    }
    return vctReturn;
}

size_t  CHLAAlignment::getcDNAsize()
{
    size_t k = 0;

     for(int i = 0; i < m_exon_sequences.size(); i++)
        k+=m_exon_sequences[i].size();

    return k;
}

vector<float> CHLAAlignment::getLeftRightScore()
{
    vector<float> vctRet;
    for(int i = 0; i < m_coverage_metrics.size(); i++)
        for(int j = 0; j < m_coverage_metrics[i].m_left_right_reads_fraction.size(); j++)
            vctRet.push_back(m_coverage_metrics[i].m_left_right_reads_fraction[j]);
    return vctRet;
}

void  CHLAAlignment::getLeftRightScore(const string& filename)
{
    int k = 0;
    ofstream out(filename.c_str());
    if(!out)
        throw(string("Can't open ")+filename);
    vector<float> score = getLeftRightScore();
    for(int i = 0; i < score.size(); i++)
    {
        //if(score[i] != -1.0f)
            out << i << '\t' << score[i] << endl;
    }
    out.close();
}

vector<long>    CHLAAlignment::getScoreFailed()
{
    long k = 0;
    vector<long> vctReturn;
    for(int i = 0; i < m_coverage_metrics.size(); i++)
    {
        for(int j = 0; j < m_coverage_metrics[i].m_central_reads_fraction.size(); j++,k++)
        {
            if(m_coverage_metrics[i].m_central_reads_fraction[j] != -1.0f)
            {
                // if(m_coverage_metrics[i].m_central_reads_fraction[j] < 0.05 ||m_coverage_metrics[i].m_left_right_reads_fraction[j] < 0.33 )
                if(m_coverage_metrics[i].m_central_reads_fraction[j] < _CHLAALIGNMENT_MIN_MIN_CENTRAL_READS_SCORE)
                    vctReturn.push_back(k);
            }
        }
    }

    return vctReturn;
}


void    CHLAAlignment::getScoreFailed(const string& filename)
{
    ofstream out(filename.c_str());
    if(!out)
        throw(string("Can't open ")+filename);
    vector<long> score = getScoreFailed();
    for(int i = 0; i < score.size(); i++)
    {
        //if(score[i] != -1.0f)
            out << score[i] << endl;
    }
    out.close();
}

vector<float>  CHLAAlignment::getCentralReadsScore()
{
    vector<float> vctRet;
    for(int i = 0; i < m_coverage_metrics.size(); i++)
        for(int j = 0; j < m_coverage_metrics[i].m_central_reads_fraction.size(); j++)
            vctRet.push_back(m_coverage_metrics[i].m_central_reads_fraction[j]);
    return vctRet;
}

void  CHLAAlignment::getCentralReadsScore(const string& filename)
{
    ofstream out(filename.c_str());
    if(!out)
        throw(string("Can't open ")+filename);
    vector<float> score = getCentralReadsScore();
    for(int i = 0; i < score.size(); i++)
    {
        //if(score[i] != -1.0f)
            out << i << '\t' << score[i] << endl;
    }
    out.close();
}

void  CHLAAlignment::getCoverage(const string& filename)
{
    ofstream out(filename.c_str());
    if(!out)
        throw(string("Can't open ")+filename);
    vector<long> coverage = getCoverage();
    for(int i = 0; i < coverage.size(); i++)
        out << i << '\t' << coverage[i] << endl;
    out.close();
}

void  CHLAAlignment::getIdealCoverage(const string& filename)
{
    ofstream out(filename.c_str());
    if(!out)
        throw(string("Can't open ")+filename);
    vector<long> coverage = getIdealCoverage();
    for(int i = 0; i < coverage.size(); i++)
        out << i << '\t' << coverage[i] << endl;
    out.close();
}

vector<long> CHLAAlignment::getExonBoundaries()  // Hier sind die Koordiunaten 0-basiert
{
    set<long> setHlp;

    for(vector<pair<unsigned int, unsigned int> >::const_iterator iter = m_exon_coordinates.begin(); iter != m_exon_coordinates.end(); iter++)
    {
        setHlp.insert(iter->first);
        setHlp.insert(iter->second);
    }
    vector<long> vctReturn;
    for(set<long>::const_iterator iter = setHlp.begin(); iter != setHlp.end(); iter++)
        vctReturn.push_back(*iter);
    return vctReturn;
}

void   CHLAAlignment::getExonBoundaries(const string& filename)
{
    ofstream out(filename.c_str());
    if(!out)
        throw(string("Can't open ")+filename);
    vector<long> boundaries = getExonBoundaries();
    for(int i = 0; i < boundaries.size(); i++)
        out << boundaries[i] << endl;
    out.close();
}

size_t  CHLAAlignment::getNumberOfCoveredBases()
{
    vector<long> cov = getCoverage();
    size_t uintRet = 0;
    for(int i = 0; i < cov.size(); i++)
        if(cov[i] > 0)
            uintRet++;
    return uintRet;
}

size_t  CHLAAlignment::getSequenceSize()
{
    size_t uintRet = 0;
    for(int i = 0; i < m_exon_sequences.size(); i++)
        uintRet+=m_exon_sequences[i].size();
    return uintRet;
}

string  CHLAAlignment::getContinuousSequence()
{
    ostringstream osr("");

    for(int i = 0; i < m_exon_sequences.size(); i++)
        osr << m_exon_sequences[i];
    return osr.str();
}

size_t CHLAAlignment::getAlignableSequenceSize(size_t minLength)
{
    size_t uintRet = 0;
    for(int i = 0; i < m_exon_sequences.size(); i++)
        if(m_exon_sequences[i].size() >= minLength)
            uintRet+=m_exon_sequences[i].size();
    return uintRet;
}

vector<CHLAAlignment::stcExonCoverage> CHLAAlignment::getExonCoverageMetrics()
{
    vector<stcExonCoverage> vctReturn;
    vector<long> cov = getCoverage();
    
    for(int i = 0; i < m_exon_coordinates.size(); i++)
    {
        const pair<unsigned int, unsigned int>& act_exon = m_exon_coordinates[i];
        size_t sum_coverage = 0;
        size_t length = act_exon.second-act_exon.first;
        stcExonCoverage act_metrics;
        act_metrics.m_central_reads_fraction.assign(length,-1.0f);
        act_metrics.m_left_right_reads_fraction.assign(length,-1.0f);
        act_metrics.maxCoverage = 0;
        act_metrics.startCoverage = cov[act_exon.first];
        act_metrics.endCoverage   = cov[act_exon.second-1];// !! Achtung, coverage ist 0-basiert
        vector<long> act_cov(length,0);
        for(unsigned int j = act_exon.first; j < act_exon.second; j++)
        {
            sum_coverage+=cov[j];
            act_cov[j-act_exon.first]=cov[j];
            if(cov[j] > act_metrics.maxCoverage)
                act_metrics.maxCoverage = cov[j];
        }
        act_metrics.avgCoverage = ((double)sum_coverage)/length;
        act_metrics.medCoverage = getMedianScore(act_cov);
        vctReturn.push_back(act_metrics);
    }
    return vctReturn;
}

void CHLAAlignment::calculateCoverageMetrics()
{
    m_coverage_metrics = getExonCoverageMetrics();
    calculateCoverageDropsExtended();
}

double CHLAAlignment::getMedianScore(vector<long> scores)
{
  double dblMedian;
  size_t size = scores.size();

  sort(scores.begin(), scores.end());

  if (size  % 2 == 0)
      dblMedian = (scores[size/2 - 1] + scores[size/2]) / 2;
  else
      dblMedian = scores[size / 2];

  return dblMedian;
}

size_t CHLAAlignment::getExonBoundaryPenalties(coverage_metrics metrics, const float factor)
{
    size_t stReturn = 0;
    
    vector<stcExonCoverage>& exon_met = m_coverage_metrics;
    double threshold = 0;

    for(int i = 0; i < exon_met.size(); i++)
    {
        switch(metrics)
        {
            case Mean:
                threshold = exon_met[i].avgCoverage*factor;
                break;
            case Median:
                threshold = exon_met[i].medCoverage*factor;
                break;
            case Max:
                threshold = exon_met[i].maxCoverage*factor;
                break;

        }
        if(exon_met[i].startCoverage < threshold)
            stReturn++;
        if(exon_met[i].endCoverage < threshold)
            stReturn++;
    }
    return stReturn;
}

pair<size_t,size_t> CHLAAlignment::countReadsAtAnchor(const vector<pair<long,long> >& ordered_reads,long anchor)
{
    pair<size_t,size_t> pReturn;
    vector<pair<long,long> > central_reads = getOverlappingReads(ordered_reads,anchor);

    for(vector<pair<long,long> >::const_iterator iter = central_reads.begin(); iter != central_reads.end(); iter++)
    {
        long L = anchor-iter->first;
        long R = iter->second-anchor;

        /*
        if(R == 0)
            pReturn.first++;
        else if(L == 0)
            pReturn.second++;
        else if(L > R && L/R > 2)
            pReturn.second++;
        else if(L >= R && L/R <= 2)
            pReturn.first++;
        else if(R > L && R/L > 2)
            pReturn.second++;
        else if(R > L && R/L <= 2)
            pReturn.first++;
         */
        if(R == 0)
            pReturn.first++;
        else if(L == 0)
            pReturn.second++;
        else if(L > R)
            pReturn.first++;
        else if(R > L)
            pReturn.second++;
        else
        {
            pReturn.first++;
            pReturn.second++;
        }

    }

    return pReturn;
}


vector<pair<long,long> > CHLAAlignment::getOverlappingReads(const vector<pair<long,long> >& ordered_reads,long pos)
{
    vector<pair<long,long> > vctReturn;
    vector<pair<long,long> >::const_iterator iter_start = ordered_reads.end();
    vector<pair<long,long> >::const_iterator iter_end = ordered_reads.end();

    for(vector<pair<long,long> >::const_iterator iter = ordered_reads.begin(); iter != ordered_reads.end() && iter_start == ordered_reads.end(); iter++)
        if(iter->first <= pos && iter->second >= pos)
            iter_start = iter;
    if(iter_start != ordered_reads.end())
    {
        for(vector<pair<long,long> >::const_iterator iter = iter_start; iter != ordered_reads.end(); iter++)
        {
            if(iter->first <= pos && iter->second >= pos)
                iter_end = iter;
            else
                break;
        }
    }
    if(iter_start != ordered_reads.end() && iter_end != ordered_reads.end())
        vctReturn.insert(vctReturn.begin(),iter_start,++iter_end);
    return vctReturn;
}

vector<size_t>  CHLAAlignment::findCoverageDropsExtended()
{
    vector<size_t> vctReturn;
    for(int i = 0; i < m_coverage_metrics.size(); i++)
    {
        for(int j = 0; j < m_coverage_metrics[i].m_central_reads_fraction.size(); j++)
        {
            if(m_coverage_metrics[i].m_central_reads_fraction[j] != -1.0f)
            {
                // if(m_coverage_metrics[i].m_central_reads_fraction[j] < 0.05 ||m_coverage_metrics[i].m_left_right_reads_fraction[j] < 0.33 )
                if(m_coverage_metrics[i].m_central_reads_fraction[j] < _CHLAALIGNMENT_MIN_MIN_CENTRAL_READS_SCORE)
                    vctReturn.push_back(j);
            }
        }
    }
    
    return vctReturn;
}

// Füge die soft clipped bereiche wieder an die reads an und shaue dir das alignment pattern an
void CHLAAlignment::calculateCoverageDropsExtended()
{
    vector<CHLAAlignment::stcAlignment> ordered_alignments = getSortedMappings();
    vector<double> area_under_curve_interpolator = getAreaUnderCurveInterpolator(); // Schlecht gecoverte Sequenzen bekommen einen Faktor auf den central reads score

    for(int i = 0; i < m_exon_coordinates.size(); i++)
    {
        long exon_start = m_exon_coordinates[i].first;     // start ist 0 basiert
        long exon_end   = m_exon_coordinates[i].second;  // end ist 0 basiert
        long exon_size  = exon_end-exon_start;

        
        if(exon_size < _CHLAALIGNMENT_MIN_EXON_LENGTH)
            continue;

        vector<CHLAAlignment::stcAlignment> overlapping_alignments = getOverlappingMappings(ordered_alignments,pair<long,long>(exon_start,exon_end));

        long exon_extend = _CHLAALIGNMENT_MAX_SOFTCLIPPING;
        if(exon_size >= _CHLAALIGNMENT_MIN_EXON_LENGTH && exon_size < _CHLAALIGNMENT_MIN_ALIGNMENT)
            exon_extend = m_read_length-exon_size;

        vector<pair<long,long> > corrected_alignments;

        for(vector<CHLAAlignment::stcAlignment>::const_iterator iter = overlapping_alignments.begin(); iter != overlapping_alignments.end(); iter++)
        {
            long the_start = (iter->RefStart+exon_extend)-iter->ReadStart;
            long the_end = the_start + m_read_length;
            corrected_alignments.push_back(pair<long,long>(the_start ,the_end));
        }

        //long start_checking = m_exon_coordinates[i].first+m_read_length/2;
        //long end_checking = m_exon_coordinates[i].second+2*exon_extend-m_read_length;
        long start_checking = max(m_exon_coordinates[i].first+m_read_length/2,(unsigned int)(m_exon_coordinates[i].first+exon_extend));
        long end_checking = min(m_exon_coordinates[i].second+2*exon_extend-m_read_length/2,m_exon_coordinates[i].second+exon_extend);

        size_t store_metrics_start = start_checking-exon_extend-exon_start;

        long virt_start = exon_start;
        long virt_end = exon_end+2*exon_extend;


        for(long j = start_checking; j < end_checking; j++)
        {
            // Korrigiere für Ungleichgewicht
            long inbalance_left = j - virt_start;
            long inbalance_right= virt_end - j;
            float correct_factor = 1.0f;
            if(inbalance_left < m_read_length && inbalance_right < m_read_length)
                correct_factor = (float)max(inbalance_right,inbalance_left)/min(inbalance_right,inbalance_left);
            else if(inbalance_left < m_read_length)
                correct_factor = (float)m_read_length/inbalance_left;
            else if(inbalance_right < m_read_length)
                correct_factor = (float)m_read_length/inbalance_right;

            pair<size_t,size_t> the_ratio_pair = countReadsAtAnchor(corrected_alignments, j);
            m_coverage_metrics[i].m_central_reads_fraction[store_metrics_start] = getCentralReadsRatio(corrected_alignments, j)*correct_factor*area_under_curve_interpolator[i];
            if( the_ratio_pair.second == 0)
                m_coverage_metrics[i].m_left_right_reads_fraction[store_metrics_start] = 0.0f;
            else
                m_coverage_metrics[i].m_left_right_reads_fraction[store_metrics_start] = ((float)the_ratio_pair.first)/((float)the_ratio_pair.second);
            store_metrics_start++;
        }
    }
}


vector<CHLAAlignment::stcAlignment> CHLAAlignment::getSortedMappings()
{
    vector<CHLAAlignment::stcAlignment> vctReturn;
    for(map<string,CHLAAlignment::stcAlignment>::const_iterator iter = m_aligned_reads.begin(); iter != m_aligned_reads.end(); iter++)
    {
        vctReturn.push_back(iter->second);
    }
    sort(vctReturn.begin(),vctReturn.end(),sort_mappings);

    //for(int i = 0; i < vctReturn.size();i++)
    //  cout << vctReturn[i].RefStart << "-" << vctReturn[i].RefStart+vctReturn[i].AlignLength << "\t" << vctReturn[i].ReadStart << endl;
    
    return vctReturn;
}

bool CHLAAlignment::sort_mappings(const CHLAAlignment::stcAlignment& one, const CHLAAlignment::stcAlignment& two)
{
    // Das reicht zum Sortieren, da Redundanzen schon beseitigt sind
    if(one.RefStart < two.RefStart)
        return true;
    else if(one.RefStart > two.RefStart)
        return false;
    else if(one.ReadStart > two.ReadStart)
        return true;
    return false;
}

vector<CHLAAlignment::stcAlignment > CHLAAlignment::getOverlappingMappings(const vector<CHLAAlignment::stcAlignment>& ordered_mappings, pair<long,long> pos)
{
    vector<CHLAAlignment::stcAlignment > vctReturn;
    vector<CHLAAlignment::stcAlignment >::const_iterator iter_start = ordered_mappings.end();
    vector<CHLAAlignment::stcAlignment >::const_iterator iter_end = ordered_mappings.end();

    for(vector<CHLAAlignment::stcAlignment>::const_iterator iter = ordered_mappings.begin(); iter != ordered_mappings.end() && iter_start == ordered_mappings.end(); iter++)
    {
        if(doesAlignmentOverlapPosition(*iter,pos))
            iter_start = iter;
    }
    if(iter_start != ordered_mappings.end())
    {
        for(vector<CHLAAlignment::stcAlignment >::const_iterator iter = iter_start; iter != ordered_mappings.end(); iter++)
            if(doesAlignmentOverlapPosition(*iter,pos))
                iter_end = iter;
            else
                break;
    }
    if(iter_start != ordered_mappings.end() && iter_end != ordered_mappings.end())
    {
        //cout << "Erster Read  - refStart: " << iter_start->RefStart << ", ReadStart: " << iter_start->ReadStart << ", AlignLength: " << iter_start->AlignLength << endl;
        //cout << "Letzter Read - refStart: " << iter_end->RefStart << ", ReadStart: " << iter_end->ReadStart << ", AlignLength: " << iter_end->AlignLength << endl;
        vctReturn.insert(vctReturn.begin(),iter_start,++iter_end);
    }
    return vctReturn;
}

// Start ist 0 basiert. Ende ist 1 basiert
bool CHLAAlignment::doesAlignmentOverlapPosition(CHLAAlignment::stcAlignment aln,pair<long,long> pos)
{
    const long& the_start = aln.RefStart;
    
    if(the_start >= pos.first && the_start  < pos.second)
        return true;
    return false;
}

// Im Idelafall sind es exakt 33 %
double CHLAAlignment::getCentralReadsRatio(vector<pair<long,long> > ordered_reads,long anchor)
{
    ordered_reads = getOverlappingReads(ordered_reads,anchor);

    double central_read = 0.0f;
    double uncentral_read = 0.0f;

    // cout << endl << "at " << anchor << endl;
    for(vector<pair<long,long> >::const_iterator iter = ordered_reads.begin(); iter != ordered_reads.end(); iter++)
    {

        long L = anchor-iter->first;
        long R = iter->second-anchor;

        if(R == 0 || L == 0)
            uncentral_read+=1.0f;
        else
        {
            double ratio = (double)L/(double)R;
            if(ratio >= 0.5 && ratio <= 2)
                central_read+=1.0f;  // central
            else
                uncentral_read+=1.0f;
        }
    }

    if(uncentral_read == 0 && central_read != 0)
        return 1.0f;
    if(uncentral_read == 0)
        return 0.0f;

    //return min(central_read,uncentral_read)/max(central_read,uncentral_read);
    return central_read/uncentral_read;
}



